/*

	jrtos.c  -  main C file containing everything

    Copyright (C) 2007  J. op den Brouw <jesse@opdenbrouw.nl>

	This file is part of jrtos version 0.1

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/


/******************************************************************
	This piece of software was written to get an idea of how to
	write an simple multitasking system on a AVR ATmega16
	microcontroller. It turned out to be quite simple once you
	get the hang of it. To learn from this software, you should
	be acquainted with the AVR microcontrollers.
	This first version can only cycle through a number of tasks.
	They are interrupted by a timer tick from TC1's overflow.
	For tracing, you can use the simulator. Though not perfect,
	it's very useful in this situation. You can also use ICE's
	like the AVR Dragon or the JTAG ICE mkII.
******************************************************************/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdint.h>

/* Prototyping */
/* Timer/Counter 1's overflow handler is used.             */
/* Naked attribute provides no register savings and return */
/* user must take care of it                               */
void TIMER1_OVF_vect( void ) __attribute__ ( ( signal, naked ) );
/* Lost tasks handler */
void catch_lost_proc(void);

/* Some defines */
#define MAX_PROCESS 8
#define STACK_LEN 64
/* Process states */
#define ACTIVE 1
#define STOPPED 2
#define EMPTY 0
/* Lost flags */
#define ALIVE 1
#define LOST 2

/* Context storage */
struct process {
	uint8_t spl;				/* Stackpointer low */
	uint8_t sph;				/* Stackpointer high */
	uint8_t stack[STACK_LEN];	/* The task's stack */
	uint8_t state;              /* The task's state, one of empty, active, stopped */
	uint8_t lost_flag;			/* Is the tast lost or not? */
	void    *pfunc;				/* The task's start address */
} context[MAX_PROCESS];

/* Pointer to currect task context */
volatile void * volatile ptr = &context[0];
/* The number of the currect task */
volatile uint8_t p=0;

/* Used by the tasks, put in here so you can see it in a watch window */
volatile uint32_t teller1=0;
volatile uint32_t teller2=0;
volatile uint32_t teller3=0;
volatile uint8_t dummy;


/* The context switcher, it's really an interrupt handler with */
/* special attributes. It first pushes the current context on  */
/* stack, handles signaling of lost tasks, then picks a new    */
/* task (simple round robin) and finally pops the new task's   */
/* context from its stack.                                     */
void TIMER1_OVF_vect( void )
{
	/* Push current context */
	/* This is politely stolen from FreeRTOS */
	asm volatile (	"push	r0						\n\t"	\
					"in		r0, __SREG__			\n\t"	\
					"cli							\n\t"	\
					"push	r0						\n\t"	\
					"push	r1						\n\t"	\
					"clr	r1						\n\t"	\
					"push	r2						\n\t"	\
					"push	r3						\n\t"	\
					"push	r4						\n\t"	\
					"push	r5						\n\t"	\
					"push	r6						\n\t"	\
					"push	r7						\n\t"	\
					"push	r8						\n\t"	\
					"push	r9						\n\t"	\
					"push	r10						\n\t"	\
					"push	r11						\n\t"	\
					"push	r12						\n\t"	\
					"push	r13						\n\t"	\
					"push	r14						\n\t"	\
					"push	r15						\n\t"	\
					"push	r16						\n\t"	\
					"push	r17						\n\t"	\
					"push	r18						\n\t"	\
					"push	r19						\n\t"	\
					"push	r20						\n\t"	\
					"push	r21						\n\t"	\
					"push	r22						\n\t"	\
					"push	r23						\n\t"	\
					"push	r24						\n\t"	\
					"push	r25						\n\t"	\
					"push	r26						\n\t"	\
					"push	r27						\n\t"	\
					"push	r28						\n\t"	\
					"push	r29						\n\t"	\
					"push	r30						\n\t"	\
					"push	r31						\n\t"	\
					"lds	r26, ptr		\n\t"	\
					"lds	r27, ptr + 1	\n\t"	\
					"in		r0, 0x3d				\n\t"	\
					"st		x+, r0					\n\t"	\
					"in		r0, 0x3e				\n\t"	\
					"st		x+, r0					\n\t"	\
				);

	/* Handle lost tasks */
	if (context[p].lost_flag == LOST) {
		context[p].state = STOPPED;
	}

	/* Select a new task to run */
	do {
		p++;
		if (p==MAX_PROCESS) { p = 0; }
	} while (context[p].state != ACTIVE);
	/* bug: if no process is active, this hangs! */

	/* Set pointer for asm routines */
	ptr = &context[p];

	/* Loads context of selected task */
	/* This is politely stolen from FreeRTOS */
	asm volatile (	"lds	r26, ptr		\n\t"	\
					"lds	r27, ptr + 1	\n\t"	\
					"ld		r28, x+					\n\t"	\
					"out	__SP_L__, r28			\n\t"	\
					"ld		r29, x+					\n\t"	\
					"out	__SP_H__, r29			\n\t"	\
					"pop	r31						\n\t"	\
					"pop	r30						\n\t"	\
					"pop	r29						\n\t"	\
					"pop	r28						\n\t"	\
					"pop	r27						\n\t"	\
					"pop	r26						\n\t"	\
					"pop	r25						\n\t"	\
					"pop	r24						\n\t"	\
					"pop	r23						\n\t"	\
					"pop	r22						\n\t"	\
					"pop	r21						\n\t"	\
					"pop	r20						\n\t"	\
					"pop	r19						\n\t"	\
					"pop	r18						\n\t"	\
					"pop	r17						\n\t"	\
					"pop	r16						\n\t"	\
					"pop	r15						\n\t"	\
					"pop	r14						\n\t"	\
					"pop	r13						\n\t"	\
					"pop	r12						\n\t"	\
					"pop	r11						\n\t"	\
					"pop	r10						\n\t"	\
					"pop	r9						\n\t"	\
					"pop	r8						\n\t"	\
					"pop	r7						\n\t"	\
					"pop	r6						\n\t"	\
					"pop	r5						\n\t"	\
					"pop	r4						\n\t"	\
					"pop	r3						\n\t"	\
					"pop	r2						\n\t"	\
					"pop	r1						\n\t"	\
					"pop	r0						\n\t"	\
					"out	__SREG__, r0			\n\t"	\
					"pop	r0						\n\t"	\
				);
	/* Let's roll..... */
	asm volatile ("reti");
}


/* The tasks!!!!!!! */
void proc0(void) {

	while (1) {
		teller1++;
	}
}
void proc1(void) {
	while (1) {
		teller2++;
	}
}
void proc2(void) {
	while (1) {
		teller3++;
	}
}

/* this one gets lost */
int proclost(void) {
	volatile uint16_t waste;

	for (waste=0; waste<1000; waste++);

	return 0;
}


/* Main routine. Is sets the context structure so tasks can run. */
int main() {

	uint8_t i;

	/* Clear contexts */
	for (i=0; i<MAX_PROCESS; i++) {
		context[i].spl=0;
		context[i].sph=0;
		context[i].state=EMPTY;
		context[i].lost_flag=ALIVE;
	}

	/* Plug in task addresses */
	context[1].pfunc = proc0;
	context[2].pfunc = proc1;
	context[3].pfunc = proc2;
	context[4].pfunc = proclost;

	/* Load the contexts. It does three things: puts the lost task catcher address     */
	/* on stack, so tasks that return get handles well; puts start adress of task      */
	/* on stack so tasks get started anyway; sets stackpointer to the right address    */
	/* so that on the task's first run, the registers are loaded (cleared) from stack; */
	/* and finally sets the task state to active.                                      */
	for (i=1; i<=4; i++) {
		context[i].stack[STACK_LEN-1] = (uint8_t)  ((uint16_t)catch_lost_proc & 0xff);
		context[i].stack[STACK_LEN-2] = (uint8_t)  (((uint16_t)catch_lost_proc >> 8) & 0xff);
		context[i].stack[STACK_LEN-3] = (uint8_t)  ((uint16_t)context[i].pfunc & 0xff);
		context[i].stack[STACK_LEN-4] = (uint8_t)  (((uint16_t)context[i].pfunc >> 8) & 0xff);
		context[i].spl = (uint8_t)  ( (uint16_t)(context[i].stack) +STACK_LEN-3-32-1-2) & 0xff;
		context[i].sph = (uint8_t) (( (uint16_t)(context[i].stack) +STACK_LEN-3-32-1-2)>>8) & 0xff;
		context[i].state=ACTIVE;
	}

	/* This routine ;-) */
	/* FIXME: the main holds context on task 0, but it doesn't use the stack */
	context[0].state=ACTIVE;

	/* Set up timer for context switches */
	TCCR1A = 0x00;
	TCCR1B = 0x01;
	TIMSK  = (1<<2);
	TIFR   = (1<<2);

	/* Free interrupts */
	sei();

	/* Do something */
	while(1) {
		dummy++;
	}

	/* Please, do NEVER return from this task */
	return 0;

}

#ifdef DEBUG_STUPID_UINT8_AND_UINT16_PROBLEM
void dummy_rout(void) {
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
	asm volatile ("nop");
}

#endif

/* This routine is called when a concurrent task executes */
/* a return instruction. It cannot return to anything,    */
/* because it was not really called. What we do is flag   */
/* the task as lost, and sit and wait until the timer     */
/* interrupt is called for a context switch.              */
void catch_lost_proc(void) {
	cli();
	context[p].lost_flag=LOST;
	sei();
	while(1);
}
